@@ -1,7 +1,8 @@
 use std::{path::PathBuf, time::Duration};
 
+use btfm_api_structs::{Clip, ClipUpdated, ClipUpload, Clips, CreatePhrase, Phrase, Phrases};
+use clap::{Parser, Subcommand};
 use reqwest::{multipart, Body, Url};
-use structopt::StructOpt;
 use thiserror::Error as ThisError;
 use tokio::fs::File;
 use tokio_util::codec::{BytesCodec, FramedRead};
@@ -9,72 +10,6 @@ use ulid::Ulid;
 
 const USER_AGENT: &str = concat!(env!("CARGO_PKG_NAME"), "/", env!("CARGO_PKG_VERSION"));
 
-// I need to factor these out and move the From impls around I think
-mod gross_hack {
-    /// Defines public-facing structures used in the web API
-    use chrono::NaiveDateTime;
-    use serde::{Deserialize, Serialize};
-
-    #[derive(Serialize)]
-    pub struct Status {
-        pub db_version: Option<u32>,
-        pub db_connections: u32,
-    }
-
-    #[derive(Clone, Debug, Serialize, Deserialize)]
-    pub struct Clip {
-        /// The unique identifier for the clip and primary key for the table.
-        pub ulid: ulid::Ulid,
-        /// The time when the clip was added to the database.
-        pub created_on: NaiveDateTime,
-        /// The last time the clip was played; this is equal to `created_on` when created.
-        pub last_played: NaiveDateTime,
-        /// Number of times the clip has been played.
-        pub plays: i64,
-        /// The output of speech-to-text on the `audio_file`, optionally used as a matching phrase.
-        pub speech_detected: String,
-        /// A description of the clip for human consumption.
-        pub description: String,
-        /// Path to the audio file, relative to the BTFM_DATA_DIR.
-        pub audio_file: String,
-        /// Phrases associated with the clip.
-        pub phrases: Option<Phrases>,
-    }
-
-    #[derive(Debug, Deserialize, Serialize)]
-    pub struct Clips {
-        pub items: u64,
-        pub clips: Vec<Clip>,
-    }
-
-    /// A phrase used to trigger one or more clips
-    #[derive(Clone, Debug, Serialize, Deserialize)]
-    pub struct Phrase {
-        pub ulid: ulid::Ulid,
-        pub phrase: String,
-    }
-
-    #[derive(Clone, Debug, Deserialize, Serialize)]
-    pub struct CreatePhrase {
-        /// The phrase.
-        pub phrase: String,
-        /// The clip to associate the phrase to.
-        pub clip: ulid::Ulid,
-    }
-
-    #[derive(Clone, Debug, Deserialize, Serialize)]
-    pub struct Phrases {
-        pub items: u64,
-        pub phrases: Vec<Phrase>,
-    }
-
-    #[derive(Debug, Deserialize, Serialize)]
-    pub struct ClipUpload {
-        pub description: String,
-        pub phrases: Option<Vec<String>>,
-    }
-}
-
 #[derive(ThisError, Debug)]
 enum Error {
     #[error("An HTTP error occurred: {0}")]
@@ -90,87 +25,97 @@ enum Error {
 /// Command-line interface for the btfm service
 ///
 /// This supports uploading, editing, and removing clips or trigger phrases.
-#[derive(StructOpt, Debug)]
-#[structopt(name = "btfm-cli")]
+#[derive(clap::Parser, Debug)]
+#[clap(name = "btfm-cli")]
+#[clap(author = "Jeremy Cline <jeremy@jcline.org>")]
+#[clap(about = "Manage clips in the BTFM Discord bot", long_about = None)]
 struct Cli {
-    #[structopt(long, env = "BTFM_URL")]
+    #[clap(long, env = "BTFM_URL")]
     url: Url,
-    #[structopt(long, short, env = "BTFM_USER")]
+    #[clap(long, short, env = "BTFM_USER")]
     user: String,
-    #[structopt(long, short, env = "BTFM_PASSWORD")]
+    #[clap(long, short, env = "BTFM_PASSWORD")]
     password: String,
-    #[structopt(subcommand)]
+    #[clap(subcommand)]
     command: Command,
 }
 
-#[derive(StructOpt, Debug)]
+#[derive(Subcommand, Debug)]
 pub enum Command {
     /// Manage audio clips for the bot
-    Clip(Clip),
+    #[clap(subcommand)]
+    Clip(ClipCommand),
     /// Manage phrases that trigger audio clips
-    Phrase(Phrase),
+    #[clap(subcommand)]
+    Phrase(PhraseCommand),
 }
 
-#[derive(StructOpt, Debug)]
-pub enum Clip {
+#[derive(Subcommand, Debug)]
+pub enum ClipCommand {
     /// Add a new clip to the database
     Add {
         /// A phrase to trigger the new clip
-        #[structopt(short, long)]
+        #[clap(short, long)]
         phrases: Option<Vec<String>>,
         /// A short description of the audio clip
-        #[structopt()]
+        #[clap()]
         description: String,
         /// The filename of the clip.
-        #[structopt(parse(from_os_str))]
+        #[clap(parse(from_os_str))]
         file: PathBuf,
     },
-    // /// Edit an existing clip in the database
-    // Edit {
-    //     /// The clip ID (from "clip list")
-    //     #[structopt()]
-    //     clip_id: Ulid,
-    //     /// A short description of the audio clip
-    //     #[structopt(short, long)]
-    //     description: Option<String>,
-    // },
+    /// Edit an existing clip in the database.
+    ///
+    /// Phrases are replaced rather than appended, so you must provide the complete list of phrases
+    /// each time you edit the clip.
+    Edit {
+        /// The clip ID (from "clip list")
+        #[clap()]
+        clip_id: Ulid,
+        /// A short description of the audio clip
+        #[clap(short, long)]
+        description: Option<String>,
+        /// The phrase or phrases that cause the clip to be played.
+        #[clap(short, long)]
+        phrases: Option<Vec<String>>,
+    },
     /// List clips in the database
     List {},
     // /// Remove clips from the database
     // Remove {
     //     /// The clip ID (from "clip list")
-    //     #[structopt()]
+    //     #[clap()]
     //     clip_id: Ulid,
     // },
 }
 
-#[derive(StructOpt, Debug)]
-pub enum Phrase {
+#[derive(Subcommand, Debug)]
+pub enum PhraseCommand {
     /// Add a trigger phrase to a clip
     Add {
         /// The clip ID (from "clip list")
-        #[structopt()]
+        #[clap()]
         clip_id: Ulid,
         /// The phrase to associate with the clip
-        #[structopt()]
+        #[clap()]
         phrase: String,
     },
     // /// Remove a phrase as a trigger for a clip
     // Remove {
     //     /// The clip ID (from "clip list")
-    //     #[structopt()]
+    //     #[clap()]
     //     clip_id: Ulid,
     //     /// The phrase ID (from "phrase list")
-    //     #[structopt()]
+    //     #[clap()]
     //     phrase_id: Ulid,
     // },
     // /// Edit an existing phrase in the database
     // Edit {
     //     /// The phrase ID (from "phrase list")
-    //     #[structopt()]
+    //     #[clap()]
     //     phrase_id: Ulid,
     //     /// The new phrase
-    //     #[structopt(short, long)]
+    //     #[clap(short, long)]
     //     phrase: String,
     // },
     /// List phrases in the database
@@ -178,14 +123,14 @@ pub enum Phrase {
     // /// Remove phrases from the database
     // Remove {
     //     /// The phrase ID (from "phrase list")
-    //     #[structopt()]
+    //     #[clap()]
     //     phrase_id: Ulid,
     // },
 }
 
 #[tokio::main(flavor = "current_thread")]
 async fn main() {
-    let opts = Cli::from_args();
+    let opts = Cli::parse();
     match process_command(opts).await {
         Ok(_) => {}
         Err(e) => eprintln!("Error: {}", e),
@@ -201,7 +146,7 @@ async fn process_command(opts: Cli) -> Result<(), Error> {
 
     match opts.command {
         Command::Clip(subcommand) => match subcommand {
-            Clip::Add {
+            ClipCommand::Add {
                 description,
                 file,
                 phrases,
@@ -218,7 +163,7 @@ async fn process_command(opts: Cli) -> Result<(), Error> {
                 let clip_part =
                     multipart::Part::stream_with_length(Body::wrap_stream(clip_stream), clip_len)
                         .file_name(file_name);
-                let clip_metadata = serde_json::to_string(&gross_hack::ClipUpload {
+                let clip_metadata = serde_json::to_string(&ClipUpload {
                     description,
                     phrases,
                 })?;
@@ -234,11 +179,11 @@ async fn process_command(opts: Cli) -> Result<(), Error> {
                     .send()
                     .await
                     .map(|resp| resp.error_for_status())??;
-                let clip = response.json::<gross_hack::Clip>().await?;
+                let clip = response.json::<Clip>().await?;
                 println!("{}", serde_json::to_string_pretty(&clip)?);
                 Ok(())
             }
-            Clip::List {} => {
+            ClipCommand::List {} => {
                 let url = opts.url.join("/v1/clips/")?;
                 let response = client
                     .get(url)
@@ -246,13 +191,36 @@ async fn process_command(opts: Cli) -> Result<(), Error> {
                     .send()
                     .await
                     .map(|resp| resp.error_for_status())??;
-                let clips = response.json::<gross_hack::Clips>().await?;
+                let clips = response.json::<Clips>().await?;
                 println!("{}", serde_json::to_string_pretty(&clips)?);
                 Ok(())
             }
+            ClipCommand::Edit {
+                clip_id,
+                description,
+                phrases,
+            } => {
+                let endpoint = format!("/v1/clips/{}", clip_id);
+                let url = opts.url.join(&endpoint)?;
+                let json = ClipUpload {
+                    description: description.unwrap_or_default(),
+                    phrases,
+                };
+                let response = client
+                    .put(url)
+                    .basic_auth(opts.user, Some(opts.password))
+                    .json(&json)
+                    .send()
+                    .await
+                    .map(|resp| resp.error_for_status())??;
+                let response = response.json::<ClipUpdated>().await?;
+                println!("{}", serde_json::to_string_pretty(&response)?);
+
+                Ok(())
+            }
         },
         Command::Phrase(subcommand) => match subcommand {
-            Phrase::List {} => {
+            PhraseCommand::List {} => {
                 let url = opts.url.join("/v1/phrases/")?;
                 let response = client
                     .get(url)
@@ -260,23 +228,23 @@ async fn process_command(opts: Cli) -> Result<(), Error> {
                     .send()
                     .await
                     .map(|resp| resp.error_for_status())??;
-                let phrases = response.json::<gross_hack::Phrases>().await?;
+                let phrases = response.json::<Phrases>().await?;
                 println!("{}", serde_json::to_string_pretty(&phrases)?);
                 Ok(())
             }
-            Phrase::Add { clip_id, phrase } => {
+            PhraseCommand::Add { clip_id, phrase } => {
                 let url = opts.url.join("/v1/phrases/")?;
                 let response = client
                     .post(url)
                     .basic_auth(opts.user, Some(opts.password))
-                    .json(&gross_hack::CreatePhrase {
+                    .json(&CreatePhrase {
                         clip: clip_id,
                         phrase,
                     })
                     .send()
                     .await
                     .map(|resp| resp.error_for_status())??;
-                let phrase = response.json::<gross_hack::Phrase>().await?;
+                let phrase = response.json::<Phrase>().await?;
                 println!("{}", serde_json::to_string_pretty(&phrase)?);
                 Ok(())
             }
